#!/usr/local/bin/perl -w

####################################################################
# Skript to create and use a webbased dictionary for the KDE
# translation
#
# kdedict Copyright (C) 1999,2000 by Matthias Kiefer <matthias.kiefer@gmx.de>
#
# kdedict is free software; you can redistribute it and/or modify	
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 or any later version.
#
# kdedict is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# you need the following perl-modules:
# DBI DBD::File Text::CSV_XS SQL::Statement
# Date::Manip
# you get them from CPAN (http://www.perl.com/CPAN)
#
# call (as root):
# perl -MCPAN -e shell
# than:
# install Date::Manip
# install DBD:File
# install Text::CSV_XS
# install SQL::Statement
####################################################################

####################################################################
# set to 0, to disable mode
# If set to 1, you dont need to log in, to maintain database.
# This is NOT recommended, when used over the web
my $MAINTAIN_MODE=0;

# the directory the database-file is in
my $DB_DIR="./kdedictionary";

# the directory the global datafiles are in:
# messages, users, languages
my $DATA_DIR=$DB_DIR;

# define Logfile for errorlogs
my $LOGFILE="$DATA_DIR/errorlogs";

# number of found entries on one page
$main::ENTRIES_IN_ONE_PAGE=15;


# set to 1, to enable multiple languages
my $MULTI_LANG_MODE=1;

# this is the name of the database. In multi-language-mode you have to use
# the string #LANG#, which is then replaced with the current language
my $TABLE="kdedictionary_#LANG#";
#my $TABLE="kdedictionary";

# this is the translated messagefile. In multi language mode
# #LANG# is replaced with language code
my $LOCALE_MSG_FILE="/home/httpd/htdocs/teams/#LANG#/messages";
#my $LOCALE_MSG_FILE="${DATA_DIR}/messages.#LANG#";
#my $LOCALE_MSG_FILE="${DATA_DIR}/messages.locale";

# this is to be able to give a default language
# set to "", if you want to get a page of available languages, if no
# parameter lang is set
my $LANGUAGE="";

####################################################################


# uncomment and change this to your local perl-directory, if you have installed
# the necessary modules in a local directory
# use lib "../PERLLIB/lib/perl5/site_perl";


####################################################################
# you shouldn't have to change anything from here
####################################################################
$main::VERSION="1.1.1";

# declare variables for use with package "strict"
use vars qw($MAINTAINER $EMAIL $TEAMPAGE $CHARSET @categories $msg_title $msg_all $msg_intro $msg_yes $msg_no
    $msg_general_error $msg_error $msg_f_date $msg_fc_date $msg_f_cat $msg_f_key $msg_f_accept $msg_f_stype $msg_f_sfield
    $msg_f_all_in_one $msg_f_begin_to_big $msg_f_begin_false $msg_access_denied $msg_fc_id
    $msg_all_entries $msg_keyword $msg_category $msg_searchtype $msg_exact $msg_begins $msg_contains
    $msg_w_search $msg_fulltext $msg_orig $msg_transl $msg_show_entries $msg_accepted $msg_n_accepted $msg_show_date
    $msg_all_in_one $msg_start $msg_new_entry
    $msg_insert_success  $msg_delete_success $msg_delete_error $msg_update_success $msg_update_error
    $msg_foot
    $msg_search_result $msg_word_head $msg_transl_head $msg_accepted_head $msg_no_entry $msg_to $msg_from
    $msg_previous_entries $msg_next_entries
    $msg_ne_entry $msg_original $msg_descrip $msg_translations
    $msg_lang_title $msg_lang_available
    $msg_is_accepted $msg_save $msg_delete $msg_help_header $msg_help_text
    $msg_login_header $msg_login_message $msg_login_user $msg_login_passwd $msg_login_login $msg_login_error $msg_login_error_lang
    $msg_user_user $msg_user_last_login $msg_user_logins $msg_user_failures $msg_user_never
    $msg_nav_login $msg_nav_logout $msg_nav_help $msg_nav_language $msg_nav_download
    $msg_wrong_sid
    $msg_pw_old $msg_pw_new $msg_pw_reenter $msg_pw_change $msg_pw_header $msg_pw_diff $msg_pw_wrong_old
    $msg_pw_too_short $msg_pw_success
    $msg_help_p1 $msg_help $msg_help_p2
    );
# the vars in last line are not used anymore but kept for compatibility reasons


# require the default messages, so we don't get an error, if new messages has been added
# but the language specific messages has not been updated, yet.
require("$DATA_DIR/messages");


# load needed modules
use strict;
use DBI;
use CGI qw/:standard :html3/;
use Date::Manip;
use Fcntl qw/:flock/;

# set environment $PATH to defined value
$ENV{'PATH'}='/bin:/usr/bin';
#enable autoflush
$|=1;
# chars, that user can search for
my $OK_CHARS='\w\-';

# a little bit more security
$CGI::POST_MAX=1000;
$CGI::DISABLE_UPLOADS=1;

alarm(60); # if the script hangs, it will be killed after 60 seconds


my $headerprinted=0; # flag,  if http-header is already send

# variables, that are used for session managment
my $expire_time; # a session expired, if the last access of the session was earlier than this
my $current_time; # the current time
my $SID=""; # the session id
my $last_login; # the date of the last login
my $total_logins; # the number of logins for the user
my $failures; # the number of failures since last login
my $user;
my $allowed_languages="";


# catch error signals to write a message to the browser and to logfile
$SIG{__DIE__} = $SIG{__WARN__}=
sub
{
   page_header();

   print "<p>$msg_general_error</p>";

   #open errorlog
   #...
   open(LOG,">>$LOGFILE") || exit(1);

   flock(LOG,LOCK_EX);

   print LOG "["
   	,UnixDate(ParseDate("today"),"%a %b %d %H:%M:%S %Y")
   	,"]: @_\n";
   close(LOG);

   page_foot();


#exit
   exit 0
};


# prepare database connection
my $DB_DSN ="DBI:CSV:f_dir=$DB_DIR";
my $DB_USER="";
my $DB_PASSWD="";

# if you change this, you have to change also the some sub functions!
my @dbfields=qw/category word_en descrip_en word_de descrip_de accept_flag last_update/;
my $dbflist =join(', ', @dbfields);

my @userdbfields=qw/user last_login total_logins failures/;
my $userdbflist =join(', ', @userdbfields);
my @sessiondbfields=qw/sid last_action user last_login total_logins failures languages ip/;
my $sessiondbflist =join(', ', @sessiondbfields);

# connect to database
my $dbh = DBI->connect($DB_DSN,$DB_USER,$DB_PASSWD) or die "Cannot connect to DB";


if($#ARGV+1 > 0)
{
   if( @ARGV->[0] eq "--version" || @ARGV->[0] eq "-v")
   {
      print << "EOT";
kdedict version $main::VERSION

kdedict Copyright (C) 1999,2000 by Matthias Kiefer <matthias.kiefer\@gmx.de>

kdedict is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 or any later version.

kdedict is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
EOT

     exit(0);
   }
   if( @ARGV->[0] eq "-init" )
   {
      init_managmentdb($dbh);

      print "Managment tables sucessfully created.\n";

      exit(0);
   }
   elsif( @ARGV->[0] eq "-init_table" )
   {
      if( $MULTI_LANG_MODE != 0 )
      {
         if($#ARGV+1 < 2)
         {
            print "Please specify a language, for which you want to have created the table.\n";
         }
         else
         {
            my $lang=@ARGV->[1];
            $TABLE =~ s/#LANG#/$lang/g;
            init_db($dbh);
            print "Translation table for language $lang successfully created.\n";
         }
      }
      else
      {
            init_db($dbh);
            print "Translation table successfully created.\n";
      }
      exit(0);
   }
}

# calculate the current and the expiring time
set_times();


if( $MULTI_LANG_MODE != 0 )
{
   if(defined param('lang'))
   {
        if( param('lang') ne "" && check_language() == 0)
   	{
      	   log_error("lang",param('lang'));
      	
           page_header();

           lang_select_page();

      	   page_foot();
      	
      	   exit 0;
   	}
   	
       $LANGUAGE=param('lang');
   }
   else
   {
      # take over default $LANGUAGE
      param('-name'=>'lang', '-value' => "$LANGUAGE");
   }

   # require translated messages
   if( $LANGUAGE ne "")
   {
      $TABLE =~ s/#LANG#/$LANGUAGE/g;
      $LOCALE_MSG_FILE =~ s/#LANG#/$LANGUAGE/g;
   }
   else
   {
       page_header();

       lang_select_page();

       page_foot();

       exit 0;
   }
}

# test if locale message file exists
-f "$LOCALE_MSG_FILE" && require("$LOCALE_MSG_FILE");


if( defined param('download') )
{
   print header('-type' => 'text/plain');

   open(TABLE,"<${DB_DIR}/$TABLE") || die "wasn't able to open table";

   flock(TABLE,LOCK_SH);

   while( <TABLE> )
   {
      print;
   }

   close(TABLE);

   exit 0;
}


if(defined param('login') )
{

   if( defined param('user') && defined param('pwd') )
   {
      my $result=check_user();
      if( $result == 0)
      {
         page_header();
         login_error_page($msg_login_error);
         log_error( "failed login: user=".param('user')." pwd=".param('pwd') );

         page_foot();
      }
      elsif ( $result == -1)
      {
         page_header();
         login_error_page($msg_login_error_lang);
         log_error( "wrong language for user: user=".param('user')." lang=".param('lang') );

         page_foot();
      }
      else
      {
         $SID = make_id();
         update_session_db();
         $MAINTAIN_MODE=1;

         page_header();
         print "<p>$msg_login_message</p>";
         print hr;

         search_form();

         page_foot();
      }
   }
   else
   {
      page_header();
      login_page();

      page_foot();
   }

   exit 0;
}

if(defined param('sid') )
{
   my $result=check_session(param('sid'));
   if( $result == 0 )
   {
      page_header();

      login_error_page($msg_wrong_sid);
      log_error( "sid",param('sid') );

      page_foot();

      exit 0;
   }
   elsif($result == -1)
   {
      page_header();

      login_error_page($msg_login_error_lang);

      read_session_db( param('sid') );
      log_error( "wrong language for session: user=$user "." lang=".param('lang') );

      page_foot();

      exit 0;
   }
   elsif($result == -2)
   {
      login_error_page($msg_wrong_sid);

      log_error( "wrong ip for valid session id: ".remote_ip() );

      page_foot();

      exit 0;
   }
   elsif( defined param('logout') )
   {
      logout( param('sid') );
   }
   else
   {
      $SID=param('sid');
      update_session_db();

      $MAINTAIN_MODE=1;
   }
}


page_header();


if(defined param('help'))
{
   help_page();
}	
elsif(param('insert') && $MAINTAIN_MODE!=0)
{
   if(param('id') !~ /^\d*$/)
   {
      my $message="${msg_fc_id}: ".param('id');
      error_page("$message");
      log_error("id",param('id'));
      goto PROGEND;
   }
   else
   {
      insert_rec($dbh,map { param($_) } @dbfields);
   }
}
elsif(param('delete') && $MAINTAIN_MODE!=0)
{
   my $id=param('id');
   if($id !~ /^\d+$/)
   {
      my $message="${msg_fc_id}: $id";
      error_page("$message");
      log_error("id",$id);
      goto PROGEND;
   }
   else
   {
      my $result=$dbh->do(<<EOT);
      DELETE FROM $TABLE
      WHERE id='$id'
EOT
   	
      $result or die "Cannot delete data";

      if($result>0)
      {
         print "<p>$msg_delete_success</p>\n";
      }
      else
      {
         log_error("id given, which is not in database: ".$id);
      print "<p>$msg_delete_error</p>\n";
   }
   }
}
elsif(param('edit') && $MAINTAIN_MODE!=0)
{
   if(defined param('id') && param('id') !~ /^\d*$/)
   {
      my $message="${msg_fc_id}: ".param('id');
      error_page("$message");
      log_error("id",param('id'));
      goto PROGEND;
   }
   else
   {
      edit_rec($dbh);
   }
}
elsif(defined param('search'))
{
   my $search_string;
   my $date;
   my $category;
   my $search_type;
   my $search_in;
   my $acceptance;

   # check date
   if(!defined param('date') || param('date') eq "")
   {
      $date="01/01/1999"; # set date before first entry
   }
   else
   {
      $date=param('date');

   	if($date !~ /^([$OK_CHARS\s]+)$/)
   	{
      	   my $message="$msg_fc_date $date";
      	   error_page("$message");
      	   log_error("date",$date);
      	   goto PROGEND;
   	}
   	else
   	{
      	   $date=ParseDate("$date");
      	   if(!$date)
      	   {
         	   error_page("$msg_f_date");
         	   goto PROGEND;
      	   }
   	}
   }

   # check categories
   if(!defined param('cat') || param('cat')eq "")
   {
      $category="$msg_all";
   }
   else
   {
      my $flag=0;
      foreach my $field ($msg_all,@categories)
      {
         if(param('cat') eq $field)
         {
            $flag=1;
            $category=param('cat');
            last;
         }
      }
      if($flag==0)
      {
         error_page("$msg_f_cat");
         log_error("cat",param('cat'));
         goto PROGEND;
      }
   }

   #check search string
   $search_string=param('search');
   if($search_string !~ /^([$OK_CHARS\s]*)$/)
   {
      my $message="$msg_f_key $search_string";
      error_page($message);
      log_error("search",$search_string);
      goto PROGEND;
   }

   #check param('af') , acceptance
   if(!defined param('af') || param('af') eq "")
   {
      $acceptance='a';
   }
   else
   {
      $acceptance=param('af');

   	if($acceptance ne 'a' && $acceptance ne 'y' && $acceptance ne 'n')
   	{
   		error_page("$msg_f_accept");
   		log_error("af",$acceptance);
   		goto PROGEND;
   	}
   }

   #check search type
   if(!defined param('type') || param('type') eq "")
   {
      $search_type='b';
   }
   else
   {
      $search_type=param('type');
      if($search_type ne 'e' && $search_type ne 'b' && $search_type ne 'c')
      {
   	   error_page("$msg_f_stype");
   	   log_error("type",$search_type);
   	   goto PROGEND;
      }
   }
   # check $search_in
   if(!defined param('in') || param('in') eq "")
   {
      $search_in='o';
   }
   else
   {
      $search_in=param('in');
      if($search_in ne 'a' && $search_in ne 'o' && $search_in ne 't')
      {
   	   error_page("$msg_f_sfield");
   	   log_error("in",$search_in);
   	   goto PROGEND;
      }
   }

   if(defined param('aio') && param('aio') ne "1")
   {
         error_page("$msg_f_all_in_one");
         log_error("aio",param('aio'));
         goto PROGEND;
   }
   # if param('b') is not a number
   if(defined param('b') && param('b')=~/\D/)
   {
       error_page("$msg_f_begin_false");
       log_error("b",param('b'));
       goto PROGEND;
   }
   elsif(!defined param('b') || param('b') eq "")
   {
      param('b',1);
   }
   # if param('m') is not a number
   if(defined param('m') && param('m')=~/\D/)
   {
       error_page("$msg_f_all_in_one");
       log_error("m",param('m'));
       goto PROGEND;
   }
   elsif(!defined param('m') || param('m')eq "")
   {
      param('m',0);
   }


   keyword_search($dbh,$search_string,$category,$search_type
	 					,$search_in,$acceptance,$date);
}
elsif(param('passwd') && $SID ne "")
{
   if(!defined param('opw') || !defined param('pw1') || !defined param('pw2') )
   {
      passwd_page();
   }
   elsif( param('pw1') ne param('pw2') )
   {
      passwd_error_page($msg_pw_diff);
   }
   else
   {
      param('user',$user);
      param('pwd',param('opw') );

      if( check_user() > 0 )
      {
         my $new_pw=param('pw1');

         if( length($new_pw) < 6 )
         {
            passwd_error_page($msg_pw_too_short);
         }
         else
         {
            change_password($user,$new_pw);

            print p("$msg_pw_success");

            print hr();
            search_form();
         }
      }
      else
      {
         passwd_error_page($msg_pw_wrong_old);
      }
   }

   page_foot();

   exit 0;
}
else
{
   print "<p>$msg_intro</p>";
}


PROGEND:
print hr();    	      	
search_form();
page_foot();

$dbh->disconnect(); # disconnect from database




#############################################
sub log_error
#############################################
{
   my ($field,$badinput)=@_;

   my $error_msg;
   if(!defined $badinput)
   {
      $error_msg=$field;
   }
   else
   {
      $error_msg="\'$badinput\' in field \'$field\'";
   }

   my $lastpage=$ENV{'HTTP_REFERER'};
   $lastpage="unknown" unless defined $lastpage;

   my $userbrowser=$ENV{'HTTP_USER_AGENT'};
   $userbrowser="unknown" unless defined $userbrowser;

   my  $useripnr=$ENV{'REMOTE_ADDR'};
   $useripnr="unknown" unless defined $useripnr;

   my $scriptname=$ENV{'SCRIPT_NAME'};
   $scriptname="unknown" unless defined $scriptname;

   open(LOG,">>$LOGFILE") || return;

   flock(LOG,LOCK_EX); # gegen parallele Zugriffe sichern

   print LOG "["
   			,UnixDate(ParseDate("today"),"%a %b %d %H:%M:%S %Y")
   			,"]:\n";
   print LOG "script: $scriptname\n",
             "browser: $userbrowser\n",
             "IP-nr: $useripnr\n",
             "from: $lastpage\n",
             "error: $error_msg\n";
   close(LOG);
}

#############################################
sub keyword_search
#############################################
{
   my ($dbh,$keyword,$category,$search_type,$search_in,$acceptance,$date)=@_;
   my $cursor;
   my $where_clause="";
   my @search_fields;
   my $keyword_low=$keyword;
   my $keyword_upp=$keyword;

   $keyword_low=~tr/-/-/;
   $keyword_upp=~tr/-/-/;

   if($search_in eq "a")
   {
      @search_fields=("word_en","word_de");
   }
   elsif($search_in eq "t")
   {
      @search_fields=("word_de");
   }
   else
   {
      @search_fields=("word_en");
   }
   if($search_type eq "b")
   {
      $keyword.='%';
      $keyword_low.='%';
      $keyword_upp.='%';
   }
   elsif($search_type eq "c")
   {
      $keyword="%$keyword%";
      $keyword_low="%$keyword_low%";
      $keyword_upp="%$keyword_upp%";

   }
   elsif($keyword eq "")
   {
      $keyword='%';
      $keyword_low='%';
      $keyword_upp='%';
   }

   if($category eq "$msg_all")
   {
      $category='%';
   }
   my $update=UnixDate($date,"%Y%m%d");

   # create sql-clause
   $where_clause="WHERE (";

   for(my $i=0;$i<=$#search_fields;$i++)
   {
      if($i!=0)
      {
         $where_clause.=" OR ";
      }
      $where_clause.="$search_fields[$i] CLIKE \'$keyword\'";
      $where_clause.=" OR ";
      $where_clause.="$search_fields[$i] CLIKE \'$keyword_low\'";
      $where_clause.=" OR ";
      $where_clause.="$search_fields[$i] CLIKE \'$keyword_upp\'";
   }
   $where_clause.=") AND category CLIKE \'$category\'";

   $where_clause.="AND last_update >= $update";

   if($acceptance eq "y")
   {
      $where_clause.=" AND accept_flag = 1";
   }
   elsif($acceptance eq "n")
   {
      $where_clause.=" AND accept_flag != 1";
   }


   my $sql=qq[ SELECT id, $dbflist
      FROM $TABLE
      $where_clause
      ORDER BY word_en];

   $cursor= $dbh->prepare($sql) or die "Select failed: $sql";

   $cursor->execute() or die "Cannot execute ($sql)";

   if(defined param('aio'))
   {
      all_in_one($cursor);
   }
   else
   {
         my $begin=param('b');
         if($begin==0)
         {
            $begin=1;
         }
         my $end=$begin+$main::ENTRIES_IN_ONE_PAGE-1;
         my $entries=param('m');
         my $output;
         my $i=0;

         if($entries<$begin)
         {
            $entries=0;
         }

         $output="<TABLE  cellpadding=4 BORDER=1 WIDTH=\"100%\">\n";
         $output.=TR(map {th($_)} ($msg_word_head, $msg_transl_head, $msg_accepted_head));

         while(defined(my $row=$cursor->fetch()))
         {
            if($i<$begin-1)
            {
                $i++;
                next;
            }
            if($i>=$end)
            {
               if($entries==0)
               {
                  $i++;
                  next;
               }
               else
               {
                  last;
               }
            }

             $output.=table_row($row);
             $i++;
          }
         $output.="</TABLE>\n";

         # if we had no Information, how many entries can be found, set them
         if($entries==0)
         {
            $entries=$i;
            param('m',$entries);
         }

         if($i==0)
         {
            $output=CGI::center(h2("$msg_search_result")).hr.p(CGI::center("$msg_no_entry"));
         }
         else
         {
            if($end>$i)
            {
               $end=$i;
               $entries=$i;

               if($begin>$end)
               {
                  error_page("$msg_f_begin_to_big");
                  log_error("b",">end: $begin");
                  goto PROGEND;
               }
            }
            my $header=CGI::center(h2("$msg_search_result"));
            $header.=CGI::center("<FONT SIZE=-1>$begin $msg_to $end $msg_from $entries</FONT>").hr;


            my $link;
            if($entries>$main::ENTRIES_IN_ONE_PAGE)
            {
               $link="<TABLE WIDTH=\"100%\"><TR>";

               if($begin>1)
               {
                  if($begin<=$main::ENTRIES_IN_ONE_PAGE)
                  {
                     param('b',1);
                  }
                  else
                  {
                     param('b',$begin-$main::ENTRIES_IN_ONE_PAGE);
                  }
                  my $self=self_url();
                  if($msg_previous_entries ne "")
                  {
                     $link.="<TD ALIGN=\"left\"><A HREF=\"$self\"><FONT SIZE=-1>$msg_previous_entries</FONT></A></TD>";
                  }
               }
               else
               {
                  $link.="<TD ALIGN=\"left\"><FONT SIZE=-1>$msg_previous_entries</FONT></TD>";
               }

               if($end<$entries)
               {
                  param('b',$begin+$main::ENTRIES_IN_ONE_PAGE);
                  my $self=self_url();
                  $link.="<TD ALIGN=\"right\"><A HREF=\"$self\"><FONT SIZE=-1>$msg_next_entries</FONT></A></TD>";
               }
               else
               {
                  $link.="<TD ALIGN=\"right\"><FONT SIZE=-1>$msg_next_entries</FONT></TD>";
               }
               $link.="</TR></TABLE>";
            }
            else
            {
               $link="";
            }
            $output=$header."<p>".$link.$output.$link."<p>";
         }

      print "$output";
  }

}

#############################################
sub all_in_one
#############################################
{
   my $cursor=shift;

   print CGI::center(h2("$msg_search_result")),hr;

   print "<p><TABLE cellpadding=4 BORDER=1 WIDTH=\"100%\">\n";
   my $i=0;
   print TR(map {th($_)} ($msg_word_head, $msg_transl_head, $msg_accepted_head));
   while(defined(my $row=$cursor->fetch()))
   {
      print table_row($row);
      $i++;
   }
   print "</TABLE>\n";

   if($i==0)
   {
      print CGI::center("$msg_no_entry");
   }
}

#############################################
sub table_row
#############################################
{
   my $row=shift;


   my $text;
   if($MAINTAIN_MODE!=0)
   {
       my $url=url();
       my $langParam="";
       if( $MULTI_LANG_MODE != 0 )
       {
          $langParam="lang=$LANGUAGE&";
       }
       $text="<a HREF=\"$url?${langParam}id=$row->[0]&edit=1";
       if($SID ne "")
       {
         $text.="&sid=$SID";
       }
       $text.="\">$row->[2]</a>";
   }
   else
   {
      $text=$row->[2];
   }

   $text="<TR><TD>$text";

   if($row->[3] ne "")
   {
      $text.="<BR>{$row->[3]}";
   }

   $text.="</td>";

   $row->[4] =~ s/\n/<BR>/g; # Newline Zeichen durch <br> ersetzen
   $text.="<td> $row->[4]";
   if($row->[5] ne "")
   {
      $text.="<BR>{$row->[5]}";
   }
   if($row->[6]==1)
   {
      $text.="<td ALIGN=CENTER WIDTH=\"10%\" BGCOLOR=\"#00FF00\">$msg_yes</td>"
   }
   else
   {
      $text.="<td ALIGN=CENTER WIDTH=\"10%\" BGCOLOR=\"#FF0000\">$msg_no</td>"
   }
   $text.="</TR>\n";

   return $text;
}


#############################################
sub insert_rec
#############################################
{
   my ($dbh,$category, $word_en, $descrip_en, $word_de, $descrip_de, $accept_flag)=@_;
   my $date=UnixDate(ParseDate("today"),"%Y%m%d");

   if($accept_flag eq $msg_yes)
   {
      $accept_flag=1;
   }
   else
   {
      $accept_flag=0;
   }

   if(param('id'))
   {
      # ID existiert => update
      my $id=param('id');
      my $sql= qq[
         UPDATE $TABLE
         SET id='$id', category='$category',
         word_en='$word_en', descrip_en='$descrip_en',
         word_de='$word_de',descrip_de='$descrip_de',
         accept_flag='$accept_flag',
         last_update='$date'
         WHERE id='$id' ];

      my $result = $dbh->do($sql);
      $result or die "Update failed ($sql)";

      if($result>0)
      {
         print "<p>$msg_update_success</p>\n";
      }
      else
      {
         print "<p>$msg_update_error</p>\n";
      }
   }
   else
   {
      # ID does not exist => new entry
      my $id=time().$$; # generate new ID
      my $sql=qq[
         INSERT INTO $TABLE
         (id,$dbflist)
         VALUES ('$id','$category', '$word_en', '$descrip_en', '$word_de', '$descrip_de', '$accept_flag','$date')
      	];
      $dbh->do($sql) or die "Insert failed ($sql)";


      print "<p>$msg_insert_success</p>\n";
   }
}

#############################################
sub edit_rec
#############################################
{
   my $dbh=shift;

   if(defined param('id'))
   {
      my $id=param('id');
      # ID exists => get field entries
      my $sql=qq[
      SELECT id, $dbflist
      FROM $TABLE
      WHERE id='$id'];
   	my $cursor=$dbh->prepare($sql) or die "Select failed: $sql";

   	$cursor->execute() or die "SQL failed";

   	while(defined(my $row=$cursor->fetch()))
   	{
            my $i=1;
      	    foreach my $field (@dbfields)
      	    {
               param($field,$row->[$i++]);
      	    }
      	    if(param('accept_flag')==1)
      	    {
      	       param('accept_flag',"$msg_yes");
      	    }
      	    else
      	    {
      	       param('accept_flag',"$msg_no");
      	    }
        }
   }

   print CGI::center(h2("$msg_ne_entry")),hr;

	
	print start_form('-method'=>'GET','-action'=>url()),
		hidden('-name' => 'id','-value'=>''),
		table({'-border'=>'1'},
			TR(td("$msg_category:"),td(popup_menu('-name'=>'category','-values'=>\@categories))),
			TR(td("$msg_original:"),td(textfield('-name'=>'word_en'))),
			TR(td("$msg_descrip:"),td(textfield('-name'=>'descrip_en'))),
			TR(td("$msg_translations:"),td(textarea('-name'=>'word_de','-rows'=>'3'))),
			TR(td("$msg_descrip:"),td(textfield('-name'=>'descrip_de'))),
			TR(td("$msg_is_accepted"),
			td(popup_menu('-name'=>"accept_flag",'-values'=>[$msg_yes,$msg_no],'default'=>"$msg_no"))),
			);
	if( $MULTI_LANG_MODE !=0 )
	{
	   print hidden('-name' => 'lang');
	}
	if($SID ne "")
	{
	   print hidden('-name' => 'sid');
	}
	print submit('-name' => 'insert', '-value' => "$msg_save");
	if( defined param('id') )
	{
      	   print submit('-name' => 'delete', '-value' => "$msg_delete"),
      	}
      	end_form();

}

#############################################
sub init_db
#############################################
{
   my $dbh = shift;
   if(! -d $DB_DIR)
   {
      mkdir($DB_DIR,0755) || die "Cannot create dir $DB_DIR";
   }

   $dbh->do(<< "EOT") or die "Cannot create table";
      CREATE TABLE $TABLE (
         id INTEGER NOT NULL,
         category VARCHAR(255),
         word_en VARCHAR(255),
         descrip_en VARCHAR(255),
         word_de VARCHAR(255),
         descrip_de VARCHAR(255),
         accept_flag INTEGER,
         last_update INTEGER
   )
EOT

}

#############################################
sub init_managmentdb
#############################################
{
   my $dbh = shift;
   if(! -d $DB_DIR)
   {
      mkdir($DB_DIR,0755) || die "Cannot create dir $DB_DIR";
   }

   $dbh->do(<< "EOT") or die "Cannot create table";
      CREATE TABLE userdb (
         user VARCHAR(255) NOT NULL,
         last_login INTEGER,
         total_logins INTEGER,
         failures INTEGER
   )
EOT

   $dbh->do(<< "EOT") or die "Cannot create table";
      CREATE TABLE sessiondb (
         sid INTEGER NOT NULL,
         last_action INTEGER,
         user VARCHAR(255),
         last_login INTEGER,
         total_logins INTEGER,
         failures INTEGER,
         languages VARCHAR(255),
         ip VARCHAR(100)
   )
EOT

}

#############################################
sub lang_select_page
#############################################
{
   my $url=url();

   open(LANGFILE,"<$DATA_DIR/languages") || die "wasn't able to open language file";

   print CGI::center(h2("$msg_lang_title")),hr;
   print "<p>$msg_intro</p>\n", hr;
   print "<p>$msg_lang_available</p><p><ul>\n";

   while( <LANGFILE> )
   {
      next if /^\s*$/; # ignore empty lines
      next if /^\s*#/; # ignore comments

      my $short; my $name;

      ($short, $name) = split( /:/ , $_);

      print "<li><a href=${url}?lang=$short>$name</a></li>\n";
   }
   print "</ul><p>\n";

   close(LANGFILE);

}

#############################################
sub help_page
#############################################
{
   print CGI::center(h2("$msg_help_header")),hr;

   print "$msg_help_text";

}

#############################################
sub login_page
#############################################
{
   print CGI::center(h2("$msg_login_header")),hr;

   login_form();      	
}

#############################################
sub login_error_page
#############################################
{
   my $msg=shift;

   print CGI::center(h2("$msg_access_denied")),hr;

   print p($msg);

   login_form();      	
}

#############################################
sub passwd_page
#############################################
{
   print CGI::center(h2("$msg_pw_header")),hr;

   passwd_form();
}

#############################################
sub passwd_error_page
#############################################
{
   my $msg=shift;

   print CGI::center(h2("$msg_pw_header")),hr;

   print p($msg);

   passwd_form();      	
}



##############################################
sub page_header
##############################################
{
   if($headerprinted==0)
   {
      # print header of page
      print header("text/html; charset=$CHARSET")
	  , start_html('-BGCOLOR' => "white", '-TITLE'=>"$msg_title")
      , h1("$msg_title");

      if(defined $SID && $SID ne "")
      {
         my $date;
         if(defined $last_login && $last_login>0)
         {
            $date=&UnixDate(&ParseDateString($last_login),"%b %e %Y %H:%M %z");
         }
         else
         {
            $date=$msg_user_never;
         }

         print << "EOT";
           <FONT SIZE=\"-1\">
           ${msg_user_user}: ${user},
           ${msg_user_last_login}: ${date},
           ${msg_user_logins}: ${total_logins}
EOT

        if($failures > 0)
        {
           print ", <FONT COLOR=\"red\"> ${msg_user_failures}: $failures</FONT>";
        }

        my $langParam="";
        my $url=url();
        if($MULTI_LANG_MODE != 0)
        {
           $langParam="lang=$LANGUAGE&";
        }
        print " - <A HREF=\"${url}?${langParam}logout=1&sid=$SID\">$msg_nav_logout</A>";


        print "</FONT>\n";

      }

      print hr();
      $headerprinted=1;
   }
}


##############################################
sub page_foot
##############################################
{
   print hr;

   my $url=url();
   my $langParam="";
   if($MULTI_LANG_MODE != 0)
   {
      $langParam="lang=$LANGUAGE&";
   }
   my $sidParam="";
   if(defined $SID && $SID ne "")
   {
      $sidParam="sid=$SID&";
   }


   print "<A HREF=\"${url}?${langParam}${sidParam}help=1\">$msg_nav_help</A> ";
   if($MULTI_LANG_MODE != 0)
   {
      print "| <A HREF=\"${url}?lang=&${sidParam}\">$msg_nav_language</A> ";
   }

   if( $MULTI_LANG_MODE == 0 || $LANGUAGE ne "")
   {
      print "| <A HREF=\"${url}?${langParam}download=1\">$msg_nav_download</A> ";

      if(defined $SID && $SID ne "")
      {
         print "| <A HREF=\"${url}?${langParam}passwd=1&sid=$SID\">$msg_pw_change</A> ";
         print "| <A HREF=\"${url}?${langParam}logout=1&sid=$SID\">$msg_nav_logout</A> ";
      }
      else
      {
         print "| <A HREF=\"${url}?${langParam}login=1\">$msg_nav_login</A>";
      }
   }

   print hr;
   print "<FONT SIZE=\"-1\">",$msg_foot,"</FONT>\n";
   print end_html();
}



#############################################
sub search_form
#############################################
{
   my $url=url();
   my $langParam="";
   if($MULTI_LANG_MODE != 0)
   {
      $langParam="lang=$LANGUAGE&";
   }

   foreach my $letter ('A'..'Z')
   {
      my $link="<a HREF=\"$url?${langParam}search=$letter";
      if($SID ne "")
      {
         $link.="&sid=$SID";
      }
      print "$link\">&nbsp;$letter</a>";
   }

    my $link="<a HREF=\"$url?${langParam}search=";
    if($SID ne "")
    {
       $link.="&sid=$SID";
    }
    print "$link\">&nbsp;$msg_all_entries</a>";


   print start_form('-method'=>'GET','-action'=>$url);
   print	"<p>$msg_keyword: ",
   		textfield('-name' => "search"),
   		"$msg_category: ",popup_menu('-name'=>"cat",'-values'=>[$msg_all,@categories],'-default'=>"$msg_all");

   print "<p>$msg_searchtype: ",
         radio_group(
         '-name' => "type",
         '-values' => ["e","b","c"],
         '-default' => "b",
         '-labels' => {"e"=>"$msg_exact ","b"=>"$msg_begins ","c"=>"$msg_contains "}
         );
   print "<p>$msg_w_search ",
         radio_group(
         '-name' => "in",
         '-values' => ["a","o","t"],
         '-default' => "o",
         '-labels' => {"a"=>"$msg_fulltext ","o"=>"$msg_orig ","t"=>"$msg_transl "}
         );

   print "<p>$msg_show_entries: ",
         radio_group(
         '-name' => "af",
         '-values' => ["y","n","a"],
         '-default' => "a",
         '-labels' => {"y"=>"$msg_accepted ","n"=>"$msg_n_accepted ","a"=>"$msg_all "}
         );

	
   print "<p>$msg_show_date ",
       textfield('-name'=>"date");	

   if($MULTI_LANG_MODE != 0)
   {
      print hidden( -name => "lang" );
   }
   if($SID ne "")
   {
      print hidden( -name => "sid" , -value => "$SID");
   }

   print	"<p>",submit('-name' => "Search", '-value' => "$msg_start");
   if($MAINTAIN_MODE!=0)
   {
      print submit('-name' => "edit", '-value' => "$msg_new_entry");
   }
   print "</p>";
    # checkbox, if we have to give the searchresults in one page
   print "<p>",
      checkbox('-name'=>'aio',
               '-value'=>'1',
               '-label'=>"$msg_all_in_one");
   print end_form();
}

#############################################
sub error_page
#############################################
{
   my $message=shift;

   page_header();

   print CGI::center(h2("$msg_error")),hr;
   print "<p>$message</p>\n";
}

#############################################
sub login_form
#############################################
{
   print "<p><center>";
   print start_form( '-method'=>'POST','-action'=>url() );
   print "<TABLE border=5 cellpadding=4 cellspacing=0>",
    TR( td({'-align'=>'right'},"$msg_login_user:"),td( textfield('-name' => "user") ) ),
    TR( td({'-align'=>'right'},"$msg_login_passwd:"),
        td( password_field('-name' => "pwd", '-override' => '1') ) )
    , "</TABLE>";

   if( $MULTI_LANG_MODE !=0 )
   {
      print hidden('-name' => 'lang');
   }

   print submit('-name' => 'login', '-value' => "$msg_login_login"),
   end_form();
   print "</center></p>"; 	
}

#############################################
sub passwd_form
#############################################
{
   print "<p><center>";
   print start_form( '-method'=>'POST','-action'=>url() );
   print "<TABLE border=5 cellpadding=4 cellspacing=0>",
    TR( td({'-align'=>'right'},"$msg_pw_old:"),
        td( password_field('-name' => "opw", '-override' => '1') ) ),
    TR( td({'-align'=>'right'},"$msg_pw_new:"),
        td( password_field('-name' => "pw1", '-override' => '1') ) ),
    TR( td({'-align'=>'right'},"$msg_pw_reenter:"),
        td( password_field('-name' => "pw2", '-override' => '1') ) )
    , "</TABLE>";

   if( $MULTI_LANG_MODE !=0 )
   {
      print hidden('-name' => 'lang');
   }

   print hidden('-name' => 'sid');

   print submit('-name' => 'passwd', '-value' => "$msg_pw_change"),
   end_form();
   print "</center></p>"; 	
}


#############################################
sub set_times
#############################################
{   	
   $current_time=UnixDate(&ParseDateString("now"),"%Y%m%d%H%M");
   $expire_time=UnixDate(&DateCalc("now","- 10minutes"),"%Y%m%d%H%M");
}

#############################################
sub check_session
#############################################
{	
   my $test_sid=shift;
   my $langs;
   my $ok=0;


   # first remove all sessions that are not valid anymore.

# This should do it, but it doesn't
#   $dbh->do(<<EOT) or die "Cannot delete data";
#      DELETE FROM sessiondb
#      WHERE last_action < $expire_time
#EOT

# so this is an ugly hack :-(
   my $sql=qq[
   SELECT sid,last_action
   FROM sessiondb];

   my $cursor=$dbh->prepare($sql) or die "Select failed: $sql";
   my $row;
   $cursor->execute() or die "SQL failed";
   while( defined($row=$cursor->fetch()) )
   {
      if($row->[1] < $expire_time)
      {
        my $sid=$row->[0];
        $dbh->do(<<EOT) or die "Cannot delete data";
        DELETE FROM sessiondb
        WHERE sid='$sid'
EOT
      }
   }


   $sql=qq[
   SELECT $sessiondbflist
   FROM sessiondb
   WHERE sid='$test_sid'];

   $cursor=$dbh->prepare($sql) or die "Select failed: $sql";

   $cursor->execute() or die "SQL failed";

   $row=$cursor->fetch();

   if( defined($row) )
   {
      # check if the language is allowed
      my $langs=$row->[6];

      my @languages=split(/,/,$langs);
      foreach (@languages)
      {
         if($_ eq $LANGUAGE || $_ eq "ALL")
         {
            $ok=1;
            last;
         }
      }
      $ok=-1 if($ok==0);


      my $test_ip=remote_ip();
      my $orig_ip=$row->[7];
      if($ok==1 && $test_ip ne $orig_ip)
      {
         $ok=-2;
      }

   }

	
   return $ok;
}


#############################################
 sub update_session_db
#############################################
{
   my $sql=qq[
   SELECT $sessiondbflist
   FROM sessiondb
   WHERE sid='$SID'];

   my $cursor=$dbh->prepare($sql) or die "Select failed: $sql";

   $cursor->execute() or die "SQL failed";

   my $row=$cursor->fetch();

   if( defined($row) )
   {
      $user=$row->[2];
      $last_login=$row->[3];
      $total_logins=$row->[4];
      $failures=$row->[5];
      my $langs=$row->[6];
      my $ip=$row->[7];

      $sql= qq[
         UPDATE sessiondb
         SET sid='$SID', last_action='$current_time',
         user='$user', last_login='$last_login',
         total_logins='$total_logins', failures='$failures',
         languages='$langs', ip='$ip'
         WHERE sid='$SID' ];

         $dbh->do($sql) or die "Update failed ($sql)";
   }
   else
   {
      my $ip=remote_ip();

      $sql= qq[
         INSERT INTO sessiondb
         ($sessiondbflist)
         VALUES ('$SID','$current_time', '$user', '$last_login', '$total_logins', '$failures', '$allowed_languages','$ip')
      	];
      $dbh->do($sql) or die "Insert failed ($sql)";
   }
}


#############################################
 sub read_session_db
#############################################
{	
   my $sid=shift;

   my $sql=qq[
   SELECT $sessiondbflist
   FROM sessiondb
   WHERE sid='$sid'];

   my $cursor=$dbh->prepare($sql) or die "Select failed: $sql";

   $cursor->execute() or die "SQL failed";

   my $row=$cursor->fetch();

   if( defined($row) )
   {
      $user=$row->[2];
      $last_login=$row->[3];
      $total_logins=$row->[4];
      $failures=$row->[5];
      my $langs=$row->[6];
      my $ip=$row->[7];
   }
   else
   {
      $user="unknown";
      $last_login="unknown";
      $total_logins="unknown";
      $failures="unknown";
   }
}


#############################################
sub make_id
#############################################
{
   my $id=unpack('H*', pack('Ncs',time(), $$ & 0xff, rand(0xffff)));

   return $id;
}

#############################################
sub remote_ip
#############################################
{
   my  $useripnr=$ENV{'REMOTE_ADDR'};
   $useripnr="0.0.0.0" unless defined $useripnr;

   return $useripnr;
}


#############################################
 sub check_user
#############################################
{
   my $test_user;
   my $passwd;
   my $langs;
   my @languages;
   my $lang;
   my $ok=0;
   $user=param('user');

   open(PWFILE, "<$DATA_DIR/users") || die "wasn't able to open user file";
 		
   flock(PWFILE,LOCK_SH);

   while(<PWFILE>)
   {		
      chomp;
      next if /^#.*$/;
      next if /^\s*$/;
 		
      if($MULTI_LANG_MODE != 0)
      {		
         ($test_user, $passwd, $langs) = split(/:/, $_);
 	
         @languages = split(/,/, $langs);			

         if($user eq $test_user)
         {
            $allowed_languages=$langs;;

            # crypt the given password
             my $test_passwd = crypt( param('pwd') , substr($passwd, 0, 2) );

             if($test_passwd eq $passwd)
             {
                foreach $lang (@languages)
                {
                   if($lang eq "ALL" || $lang eq $LANGUAGE)
                   {
                      $ok=1;
                      last;
                   }
                }

                $ok = -1 if( $ok==0 );
             }

             last;
         }
      }
      else
      {   	
         ($test_user, $passwd) = split(/:/, $_);
 	
         if($user eq $test_user)
         {
            $allowed_languages="ALL";

            # crypt the given password
            my $test_passwd = crypt( param('pwd') , substr($passwd, 0, 2) );

             if($test_passwd eq $passwd)
             {
                $ok=1;
             }

             last;
         }
      }
   }

   close(PWFILE);

   # update the users information in the user database
   my $sql=qq[
   SELECT $userdbflist
   FROM userdb
   WHERE user='$user'];

   my $cursor=$dbh->prepare($sql) or die "Select failed: $sql";

   $cursor->execute() or die "SQL failed";

   my $row=$cursor->fetch();

   if( !defined($row) )
   {
      if($ok==1)
      {
         $last_login=0;
         $total_logins=$ok;
         $failures=0;
         $failures=1 if $ok==0;

         my $login_time=0;
         $login_time=$current_time if($ok==1);

         $sql=qq[
         INSERT INTO userdb
         ($userdbflist)
         VALUES ('$user', '$login_time', '$total_logins' ,'$failures')
         ];

         $dbh->do($sql) or die "Insert failed ($sql)";
      }
   }
   else
   {
      $last_login=$row->[1];
      $total_logins=$row->[2];
      $failures=$row->[3];

      my $login_time=$last_login;
      my $new_failures=$failures;

      if($ok==1)
      {
         $total_logins++;
         $login_time=$current_time;
         $new_failures=0;
      }
      else
      {
         $new_failures++;
      }

      my $sql= qq[
         UPDATE userdb
         SET user='$user', last_login='$login_time',
         total_logins='$total_logins', failures='$new_failures'
         WHERE user='$user' ];

         $dbh->do($sql) or die "Update failed ($sql)";

   }

   return $ok;
}


#############################################
 sub check_language
#############################################
{
   my $test_lang=param('lang');
   my $ok=0;

   open(LANGFILE,"<$DATA_DIR/languages") || die "wasn't able to open language file";

   while( <LANGFILE> )
   {
      chomp;
      next if /^\s*$/; # ignore empty lines
      next if /^\s*#/; # ignore comments

      my $short; my $name;

      ($short, $name) = split( /:/ , $_);

      if($short eq $test_lang)
      {
         $ok=1;
         last;
      }
   }

   close(LANGFILE);

   return $ok;
}

#############################################
sub logout
#############################################
{
    my $sid = shift;

    $dbh->do(<<EOT) or die "Cannot delete data";
    DELETE FROM sessiondb
    WHERE sid='$sid'
EOT

}


#############################################
sub change_password
#############################################
{
   my $change_user=shift;
   my $new_pw=shift;

   my $crypt_pw = crypt( $new_pw , substr( time() , 0, 2) );

   open(PWFILE, "+<${DATA_DIR}/users") || die "wasn't able to open user file"; 		

   flock(PWFILE,LOCK_EX);

   my $contents="";
   my $done=0;
   while(<PWFILE>)
   {	
      if($done==1 || /^#.*$/ || /^\s*$/)
      {
         $contents.=$_;
         next;
      }

      my $tmp_user;
      my $passwd;
      ($tmp_user, $passwd) = split(/:/, $_);

      if( $tmp_user eq $change_user )
      {
         # replace old password
         s/:$passwd/:$crypt_pw/;

         chomp;
         $contents.=($_."\n");

         $done=1;
      }
      else
      {
         $contents.=$_;
      }
   }

   ($done == 1) || die "wasn't able to find user in userlist";

   # now delete the old file and and write the new contents
   seek(PWFILE,0,0);
   truncate(PWFILE,0);

   print PWFILE "$contents";

   close(PWFILE);
}
